#SkotBoot - CS150 MIPS bootloader
	## sends receives hex file as ASCII
	## loads it into imem or dmem
.text
	.globl  _start
	.align  2
	.type   _start,@function
	.ent    _start
_start:
	.set    noreorder

	# setup $sp to start at 0x1001FFFC
	lui $sp, 0x1001
	j   main
	ori $sp, $sp, 0xFFFC

##fauxISR -  will jump to this location in bootloaded code
	la $k0, runprogram
	addi $k0, $k0, 0xC
	jr $k0
	nop

#sends $a0 to uart - doesn't hurt $a0
#leaf method
printbyte:
	lui $t0, 0xFFFF			#load address of UART0DataInReady into $t0
	ori $t0, $t0, 0x0008
	lw $t1, 0($t0)			#get status of UART0DataInReady
	beq $t1, $0, printbyte		#loop if UART0DataInReady == 0
	nop
	lui $t2, 0xFFFF			#load address of UART0DataIn
	ori $t2, $t2, 0x000c
	sw $a0, 0($t2)			#save byte to UART0DataIn

	jr $ra
	nop
#end printbyte function
	
#returns byte in $v0
getbyte:
	lui $t0, 0xFFFF			#load address of UART0DataOutValid into $t0
	ori $t0, $t0, 0x0000
	lw $t1, 0($t0)			#get status of UART0DataOutValid
	beq $t1, $0, getbyte		#loop if UART0DataOutValid == 0
	nop

	lui $t2, 0xFFFF			#load address of UART0DataOut
	ori $t2, $t2, 0x0004
	lw $v0, 0($t2)			#load byte from UART0DataOut
	jr $ra
	nop
#end getbyte function

#takes an ascii char in $a0
#returns it's hex value in $v0
asciitohex:

	#ex: 0x0000_000a would be: 0x0000_0061 or 0x0000_0041

	#test to see if it is hex
	addi $t1, $0, 0x2f		# one below '0'
	slt $t0, $t1, $a0		#set $t0 if $t1 < $a0
	beq $t0, $0, ascii_not_hex	#$a0 not printable
	nop
	
	#test to see if it is a number
	addi $t1, $0, 0x39		#9
	slt $t0, $t1, $a0		#set $t0 if $t1 < $a0
	beq $t0, $0, ascii_number	#$a0 is a number	
	nop
	
	#test to see if it is hex
	addi $t1, $0, 0x40		#one below 'A'
	slt $t0, $t1, $a0		#set $t0 if $t1 < $a0
	beq $t0, $0, ascii_not_hex	#$a0 not printable
	nop
	
	#test to see if it is an uppercase letter
	addi $t1, $0, 0x46		#'F'
	slt $t0, $t1, $a0		#set $t0 if $t1 < $a0
	beq $t0, $0, ascii_uletter	#$a0 is a uletter
	nop
	
	#test to see if it is hex
	addi $t1, $0, 0x60		#one below 'a'
	slt $t0, $t1, $a0		#set $t0 if $t1 < $a0
	beq $t0, $0, ascii_not_hex	#$a0 not printable
	nop
	
	#test to see if it is a lowercase letter
	addi $t1, $0, 0x66		#'f'
	slt $t0, $t1, $a0		#set $t0 if $t1 < $a0
	beq $t0, $0, ascii_letter	#$a0 is a letter
	nop

	#if we get here it is not hex.. roll through to ascii_not_hex
	
ascii_not_hex:
	#make it a zero
	add $v0, $0, $0
	j asciitohex_done
	nop

ascii_number:
	addi $v0, $a0, -48
	j asciitohex_done
	nop
	
ascii_uletter:
	addi $v0, $a0, -55
	j asciitohex_done
	nop
	
ascii_letter:
	addi $v0, $a0, -87
	
asciitohex_done:	
	jr $ra
	nop
#end asciitohex function
	

#calls getbyte 8 times
#print each byte received
#convert ascii to hex
#returns a full word in $v0
#sets $v1 if eot
getword:
	#running tally in $s0
	
	addi $sp, $sp, -8
	sw $s0, 4($sp)
	sw $ra, 0($sp)

	#get ascii char 1 - most significant 4 bits
	jal getbyte			#getbyte returns a byte in $v0
	nop
	
	#the first byte could be eot	
	addi $t1, $0, 0x04		#0x04 = EOT
	beq $v0, $t1, getword_eot	#if it is a eot, branch to getword_eot and set $v1
	nop

	add $a0, $0, $v0		#move $v0 to $a0 for printbyte
	jal printbyte			#printbyte doesn't hurt $a0
	nop

	#it's not eot, convert to ascii. takes $a0, returns $v0
	jal asciitohex
	nop

	#shift the result into the most significant 4 bits
	sll $t0, $v0, 28
	add $s0, $0, $t0

	## loop to shift char 2-8
	addi $t3, $0, 28		#init counter in $t3
	
getwordloop:
	addi $t3, $t3, -4		#decrement counter
	
	jal getbyte
	nop
	add $a0, $0, $v0
	jal printbyte
	nop
	jal asciitohex
	nop
	sllv $t0, $v0, $t3
	or $s0, $s0, $t0
	
	bne $t3, $0, getwordloop 	#loop back
	nop
	## end getwordloop
	
	#get ascii char 9 (newline)
	jal getbyte			#print newlines, but don't do anything else..
	nop
	add $a0, $0, $v0
	jal printbyte
	nop

	add $v0, $0, $s0		#move $s0 into $v0 to return full word
	add $v1, $0, $0			#not eot
	
	lw $ra, 0($sp)
	lw $s0, 4($sp)
	addi $sp, $sp, 4
	
	jr $ra
	nop

getword_eot:
	lw $ra, 0($sp)
	lw $s0, 4($sp)
	addi $sp, $sp, 4
	
	addi $v1, $0, 1			#not eot
	jr $ra
	nop
#end getword function
	
loadimem:
#s0 is a memory address counter
#s1 is the base memory address

	add $s0, $0, $0			#init $s0

	la $s1, runprogram		#get the address of the end of the bootloader

loadimem_loop:	
	jal getword
	nop
	addi $t0, $0, 1
	beq $v1, $t0, main		#jump back to main if it is eot
	nop
	
	#else load the word into imem!
	add $t1, $s0, $s1		#get memory address and put in $t1
	sw $v0, 0($t1)

	addi $s0, $s0, 4		#increment memory counter
	
	j loadimem_loop
	nop
#end loadimem function

loaddmem:
#s0 is a memory address counter
#s1 is the base memory address

	add $s0, $0, $0			#init $s0
	
	lui $s1, 0x1001			#data memory starts at 0x1001_0000
	ori $s1, $s1, 0x0000

loaddmem_loop:	
	jal getword
	nop
	addi $t0, $0, 1
	beq $v1, $t0, main		#jump back to main if it is eot
	nop
	
	#else load the word into dmem!
	add $t1, $s0, $s1		#get memory address and put in $t1
	sw $v0, 0($t1)

	addi $s0, $s0, 4		#increment memory counter
	
	j loaddmem_loop
	nop
#end loaddmem function

####
############# MAIN
main:
	## sendwelcome
	addi $a0, $0, 0x0a	#newline
	jal printbyte
	nop
	
	addi $a0, $0, 0x3f	#?
	jal printbyte
	nop

	addi $a0, $0, 0x3e	#>
	jal printbyte
	nop
	
	jal getbyte
	nop
	
	addi $t0, $0, 0x69		#i - Load imem
	beq $v0, $t0, loadimem
	nop
	
	addi $t0, $0, 0x64		#d - Load dmem
	beq $v0, $t0, loaddmem
	nop
	
	addi $t0, $0, 0x72		#r - Run Program
	beq $v0, $t0, runprogram
	nop
	j main				#invalid character, sendwelcome again so they get the hint..
	nop

#the cliff
runprogram:
	nop				#this is where imem will be written to	
	nop				
	nop
	nop				#this is where the faux ISR jumps to.
	#bon voyage!

.end _start
